@@ -68,6 +68,8 @@ pkgCacheGenerator::pkgCacheGenerator(DynamicMMap *pMap,OpProgress *Prog) :
 		    Map(*pMap), Cache(pMap,false), Progress(Prog),
 		     CurrentRlsFile(nullptr), CurrentFile(nullptr), d(nullptr)
 {
+   // Reserve some space for version extra data, to avoid lots of small resizes.
+   VersionExtra.reserve(32 * 1024);
 }
 bool pkgCacheGenerator::Start()
 {
@@ -249,6 +251,15 @@ bool pkgCacheGenerator::MergeList(ListParser &List,
       if (PackageName.empty() == true)
 	 return false;
 
+      // Package is excluded
+      if (exclude.find(PackageName) != exclude.end())
+	 continue;
+
+      // Includes are specified, this is the total set of packages to include,
+      // so we need to skip the package if it is not in the include list.
+      if (not include.empty() && include.find(PackageName) == include.end())
+	 continue;
+
       Counter++;
       if (Counter % 100 == 0 && Progress != 0)
 	 Progress->Progress(List.Offset());
@@ -396,8 +407,8 @@ bool pkgCacheGenerator::MergeListVersion(ListParser &List, pkgCache::PkgIterator
 	    if (List.SameVersion(Hash, Ver))
 	    {
                // We do not have SHA256 for both, so we cannot compare them, trust the call from SameVersion()
-               if (ListSHA256.empty() || VersionExtra[Ver->ID].SHA256[0] == 0)
-                  break;
+	       if (ListSHA256.empty() || Ver->ID >= VersionExtra.size() || VersionExtra[Ver->ID].SHA256[0] == 0)
+		  break;
                // We have SHA256 for both, so they must match.
                if (ListSHA256 == std::string_view(VersionExtra[Ver->ID].SHA256, 64))
                   break;
@@ -1330,6 +1341,7 @@ bool pkgCacheGenerator::SelectFile(std::string const &File,
 				   pkgIndexFile const &Index,
 				   std::string const &Architecture,
 				   std::string const &Component,
+				   const IndexTarget *Target,
 				   unsigned long const Flags)
 {
    CurrentFile = nullptr;
@@ -1372,6 +1384,22 @@ bool pkgCacheGenerator::SelectFile(std::string const &File,
    Cache.HeaderP->FileList = map_pointer<pkgCache::PackageFile>{NarrowOffset(CurrentFile - Cache.PkgFileP)};
    Cache.HeaderP->PackageFileCount++;
 
+   include.clear();
+   exclude.clear();
+   if (Target)
+   {
+      if (auto inc = Target->Options.find("INCLUDE"); inc != Target->Options.end())
+      {
+	 auto v = VectorizeString(inc->second, ' ');
+	 include.insert(v.begin(), v.end());
+      }
+      if (auto exc = Target->Options.find("EXCLUDE"); exc != Target->Options.end())
+      {
+	 auto v = VectorizeString(exc->second, ' ');
+	 exclude.insert(v.begin(), v.end());
+      }
+   }
+
    if (Progress != 0)
       Progress->SubProgress(Index.Size());
    return true;
